import getpass
import importlib
import os
import runpy
import sys
from contextlib import contextmanager

import pytest
from clifire import application, main

from tests.test_output import output


def get_app():
    module_name = 'tests.sample.fire'
    if module_name in sys.modules:
        del sys.modules[module_name]
    app = application.App()
    importlib.import_module(module_name)
    return app


def test_main_block(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', ['clifire/main.py', 'help'])
    if 'clifire.main' in sys.modules:
        del sys.modules['clifire.main']
    runpy.run_module('clifire.main', run_name='__main__')
    assert 'Show this help' in output(capsys)


def test_fire_command(capsys):
    app = get_app()
    app.fire('hello Rob')
    assert 'Hi Rob!' in output(capsys)
    app.fire('hello Rob -v')
    assert 'Hi Rob!' in output(capsys)
    app.fire('hello -v Rob')
    assert 'Hi Rob!' in output(capsys)
    app.fire('-v hello Rob')
    assert 'Hi Rob!' in output(capsys)
    app.fire('hello')
    assert f'Hi {getpass.getuser()}!' in output(capsys)
    app.fire('hello "Rob Lizana"')
    printed = output(capsys)
    assert 'Hi Rob Lizana!' in printed
    assert 'Sudo = False' in printed
    app.fire('hello -v "Rob Lizana" --sudo')
    printed = output(capsys)
    assert 'Hi Rob Lizana!' in printed
    assert 'Sudo = True' in printed


def test_fire_without_doc(capsys):
    app = get_app()
    app.fire('nodoc')
    assert 'Command without doc' in output(capsys)


@contextmanager
def in_path(*args):
    cwd = os.getcwd()
    os.chdir(application.App.path(os.path.dirname(__file__), *args))
    try:
        yield
    finally:
        os.chdir(cwd)


def test_fire_main(capsys):
    with in_path('sample'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)


def test_fire_not_found_in_parents(capsys):
    temp_dir = os.getenv('TMPDIR') or '/tmp'
    assert os.path.isdir(temp_dir)
    with in_path(temp_dir):
        with pytest.raises(SystemExit):
            main.main('nodoc')
        assert 'Command without doc' not in output(capsys)


def test_fire_find_file_py_in_parents(capsys):
    with in_path('sample', 'parent'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)
    with in_path('sample', 'parent_fire', 'child'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)
    with in_path('sample', 'parent_fire', 'child', 'grandson'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)


def test_fire_find_folder_in_parents(capsys):
    with in_path('sample', 'parent_fire'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)
    with in_path('sample', 'parent_fire', 'child'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)
    with in_path('sample', 'parent_fire', 'child', 'grandson'):
        main.main('nodoc')
        assert 'Command without doc' in output(capsys)


def test_fire_group_help(capsys):
    app = get_app()

    app.fire('help')
    printed = output(capsys)
    assert 'Available Commands:' in printed
    assert ' cd ' in printed
    assert ' ef ' in printed
    assert ' ef gh ' in printed
    assert 'doc_ab_cd' in printed
    assert 'doc_ab_ef_gh' in printed

    app.fire('help ab')
    printed = output(capsys)
    assert 'Available Commands:' in printed
    assert 'doc_ab' in printed
    assert 'doc_ab_cd' in printed
    assert 'ef gh' in printed
    assert 'doc_ab_ef_gh' in printed

    app.fire('help ab ef')
    printed = output(capsys)
    assert 'ab [options]' in printed
    assert 'ab ef [options]' not in printed
    assert 'doc_ab_ef_gh' in printed

    app.fire('help zz')
    assert 'Available Commands:' in output(capsys)

    app.fire('help zz command')
    assert 'Available Commands:' not in output(capsys)

    app.fire('ab')
    assert 'def_ab' in output(capsys)

    app.fire('ab cd')
    assert 'def_ab_cd' in output(capsys)

    app.fire('ab ef gh')
    assert 'def_ab_ef_gh' in output(capsys)

    app.fire('zz command')
    assert 'def_zz_command' in output(capsys)



import re

from clifire import command, out


@command.fire
def update_version(cmd):
    '''
    Update the clifire version according to pyproject.toml
    '''
    pyproject_path = cmd.app.path('pyproject.toml')
    with open(pyproject_path) as f:
        content = f.read()
    match = re.search(r'version\s*=\s*["\']([^"\']+)["\']', content)
    if match is None:
        out.critical(f'Version not detected {pyproject_path}')
    version = match.group(1)
    out.debug(f'Version "{version}" in {pyproject_path}')
    with open(cmd.app.path('src/clifire/main.py')) as f:
        content = f.read()
    new_content = re.sub(
        r'app = application\.App\('
        r'name=["\'"]CliFire["\'], '
        r'version=["\'"][^\'"]+["\']\)',
        f'app = application.App(name=\'CliFire\', version=\'{version}\')',
        content,
    )
    if new_content != content:
        with open(cmd.app.path('src/clifire/main.py'), 'w') as f:
            f.write(new_content)
        out.success(f'Update the version to {version}')
    else:
        out.warn(f'The version is correct, version {version}')



from clifire import application, command, out


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_command_help(capsys):
    app = application.App(name='sample', version='1.0.99')
    app.fire('help')
    printed = output(capsys)
    assert 'Description' in printed
    assert 'Arguments' in printed
    assert 'version' in printed
    assert 'help' in printed


def test_command_help_without_help(capsys):
    class CommandTest(command.Command):
        _name = 'test_no_help'

    app = application.App()
    app.add_command(CommandTest)
    cmd = app.get_command('test_no_help')
    assert cmd._name == 'test_no_help'
    assert cmd._help is None
    app.fire('help test_no_help')
    printed = output(capsys)
    assert 'test_no_help' in printed


def test_command_help_without_arguments(capsys):
    class CommandTest(command.Command):
        _name = 'test'
        _help = 'Test sample'

    app = application.App(name='sample', version='1.0.99')
    app.add_command(CommandTest)
    app.fire('help test')
    printed = output(capsys)
    assert 'Description' in printed
    assert 'Arguments' not in printed


def test_command_help_without_global_options(capsys):
    app = application.App(name='sample', version='1.0.99')
    app.fire('help')
    printed = output(capsys)
    assert '--no-ansi' in printed
    assert '--verbose' in printed

    app = application.App(
        name='sample',
        version='1.0.99',
        option_verbose=False,
        option_ansi=False,
    )
    app.fire('help')
    printed = output(capsys)
    assert 'Description' in printed
    assert 'Global options:' in printed
    assert '--no-ansi' not in printed
    assert '--verbose' not in printed
    assert 'Arguments' in printed
    assert 'version' in printed
    assert 'help' in printed

    app.options = {}
    app.fire('help')
    printed = output(capsys)
    assert 'Global options:' not in printed


def test_command_help_with_groups(capsys):
    class CommandDbCreate(command.Command):
        _name = 'db.create'
        _help = 'Create database'

    class CommandDbDrop(command.Command):
        _name = 'db.drop'
        _help = 'Drop database'

    class CommandConfigSet(command.Command):
        _name = 'config.set'
        _help = 'Set config value'

    app = application.App(name='sample', version='1.0.99')
    app.add_commands(
        [
            CommandDbCreate,
            CommandDbDrop,
            CommandConfigSet,
        ]
    )

    app.fire('help')
    printed = output(capsys)
    assert 'db' in printed
    assert 'config' in printed
    assert 'create' in printed
    assert 'drop' in printed
    assert 'set' in printed
    assert 'Create database' in printed
    assert 'Drop database' in printed
    assert 'Set config value' in printed

    db_pos = printed.find('db')
    config_pos = printed.find('config')
    assert config_pos < db_pos

    app.fire('db create --help')
    printed = output(capsys)
    assert 'Create database' in printed

    app.fire('db create -h')
    printed = output(capsys)
    assert 'Create database' in printed



from clifire import command, out


class CommandVersion(command.Command):
    _name = 'version'
    _help = 'Display project version'

    def fire(self):
        out.info(f'{self.app.name} {self.app.version}')



import os
import shlex
import subprocess
import sys
from typing import Any, Dict, List, Type

from clifire import command, commands, config, out, result, template


class App:
    current_app = None

    def __init__(
        self,
        name: str = '',
        version: str = '0.0.1 alpha',
        context: Dict[str, Any] = None,
        option_verbose: bool = True,
        option_ansi: bool = True,
        config_files: List[str] = None,
        config_create: bool = False,
        command_help=commands.help.CommandHelp,
        command_version=commands.version.CommandVersion,
        template_folder=None,
        show_messages_with_icons: bool = True,
    ):
        App.current_app = self
        self.name = name
        self.version = version
        self.options = {}
        if option_verbose:
            self._add_option_verbose()
        if option_ansi:
            self._add_option_ansi()
        if context is None:
            context = {}
        self.context = context
        if config_files is None:
            config_files = []
        self.config = config.Config.get_config(
            config_files, create=config_create, **context
        )
        self.commands = {}
        if command_help:
            self.add_option(
                'help',
                command.Field(
                    help='Show help',
                    default=False,
                    alias='h',
                ),
            )
            self.add_command(command_help)
        if command_version:
            self.add_command(command_version)
        self.template = None
        if template_folder:
            self.template = template.Template(template_folder)
        self.show_messages_with_icons = show_messages_with_icons

    def _add_option_verbose(self):
        self.add_option(
            'verbose',
            command.Field(
                help='Verbose mode',
                default=False,
                alias='v',
            ),
        )
        command_line = sys.argv[1:]
        if '-v' in command_line or '--verbose' in command_line:
            self.set_option('verbose', True)

    def _add_option_ansi(self):
        self.add_option(
            'no_ansi',
            command.Field(
                help='Disable colored output',
                default=False,
            ),
        )
        command_line = sys.argv[1:]
        if '--no-ansi' in command_line:
            self.set_option('no_ansi', True)

    def add_option(self, name: str, field: command.Field):
        self.options[name] = field
        for alias in field.alias:
            if alias.startswith('-'):
                alias = alias[2:] if alias.startswith('--') else alias[1:]
            alias = alias.replace('-', '_')
            if alias in self.options:
                raise command.CommandException(
                    f'Duplicate global option alias "{alias}"'
                )
            self.options[alias] = name

    def set_option(self, name: str, value):
        if name not in self.options:
            return False
        self.options[name].value = value
        return True

    def get_option(self, name: str, default=None):
        if name not in self.options:
            return default
        return self.options[name].value

    def add_command(self, cls: Type[command.Command]):
        if not cls._name:
            raise command.CommandException(
                f'The command {cls} has no name, please set _name var in class'
            )
        self.commands[cls._name] = cls

    def add_commands(self, commands: List[Type[command.Command]]):
        for cmd in commands:
            self.add_command(cmd)

    def _split_command_line(self, command_line: str) -> list:
        params = shlex.split(command_line)
        args = [p for p in params if not p.startswith('-')]
        if 'help' in self.commands and 'help' not in params:
            if '--help' in params:
                params.remove('--help')
                params.insert(0, 'help')
            if '-h' in params:
                params.remove('-h')
                params.insert(0, 'help')
            if not args:
                params.insert(0, 'help')
        if '-v' in params or '--verbose' in params:
            self.set_option('verbose', True)
        if '--no-ansi' in params:
            self.set_option('no_ansi', True)
        return params

    def _clean_command_line(self, command_line: str) -> str:
        return shlex.join(self._split_command_line(command_line))

    def find_command(self, command_line: str):
        empty = False
        args = [
            p
            for p in self._split_command_line(command_line)
            if not p.startswith('-')
        ]
        if not args:
            args = ['help']
            empty = True
        while args:
            command_name = '.'.join(args)
            if command_name in self.commands:
                return self.commands[command_name]
            args.pop()
        if empty:
            out.critical('No command provided.', code=10)
        return None

    def get_command(self, command_line: str) -> command.Command:
        command_line = self._clean_command_line(command_line)
        cls = self.find_command(command_line)
        if cls:
            return cls(self, command_line)
        args = [p for p in shlex.split(command_line) if not p.startswith('-')]
        while args:
            group = '.'.join(args + [''])
            commands = [k for k in self.commands.keys() if k.startswith(group)]
            if commands:
                cls = self.find_command('help')
                if cls:
                    return cls(self, f'help {command_line}')
            last_arg = args.pop()
        out.critical(f'Command "{last_arg}" not found.', code=20)

    def fire(self, command_line: str = None):
        try:
            if command_line is None:
                out.debug(f'Sys argv value: {sys.argv}')
                command_line = shlex.join(sys.argv[1:])
            else:
                command_line = self._clean_command_line(command_line)
            cmd = self.get_command(command_line)
            out.setup(
                not self.get_option('no_ansi'),
                show_icons=self.show_messages_with_icons,
            )
            res = cmd.launch(cmd.command_line)
            if type(res) is int and res != 0:
                sys.exit(res)
        except command.CommandException as e:
            out.critical(e, code=30)
        except command.FieldException as e:
            out.critical(e, code=40)
        except KeyboardInterrupt:
            out.error('Keyboard interrupt!')
            raise

    @classmethod
    def shell(
        cls,
        cmd: str,
        capture_output: bool = True,
        env: dict = None,
        path: str = None,
        shell: bool = True,
        check: bool = False,
    ) -> result.Result:
        pwd = os.getcwd()
        try:
            if path:
                out.debug(f'Shell path: {path}')
                os.chdir(path)
            env_vars = os.environ.copy()
            if env:
                env_vars.update(env)
            out.debug(f'Shell: {cmd}')
            proc = subprocess.run(
                cmd if shell else shlex.split(cmd),
                shell=shell,
                capture_output=capture_output,
                env=env_vars,
                check=check,
            )
            return result.Result(proc.returncode, proc.stdout, proc.stderr)
        except subprocess.CalledProcessError as e:
            return result.ResultError(e.stderr, e.returncode)
        finally:
            os.chdir(pwd)

    @classmethod
    def path(cls, *args: List[str]) -> str:
        if len(args) == 0:
            args = (os.getcwd(),)
        exapnd_path = os.path.join(
            *(a.replace('~', os.path.expanduser('~')) for a in args)
        )
        return os.path.abspath(exapnd_path)



import os
import tempfile

import pytest
import yaml
from clifire import config


@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield tmpdirname


def test_config(temp_dir):
    config_file = os.path.join(temp_dir, 'config.yaml')
    test_data = {
        'name': 'test',
        'value': 123,
        'list': ['item1', 'item2'],
        '_private': 'hidden',
    }

    conf = config.Config(config_file=config_file)
    assert not conf.read()

    conf.name = test_data['name']
    conf.value = test_data['value']
    conf.list = test_data['list']
    conf._private = test_data['_private']
    conf.write()
    assert os.path.exists(config_file)
    with open(config_file, encoding='utf-8') as f:
        saved_data = yaml.safe_load(f)
    assert 'name' in saved_data
    assert '_private' not in saved_data

    conf = config.Config(config_file=config_file)
    assert conf.read()
    assert conf.name == test_data['name']
    assert conf.value == test_data['value']
    assert conf.list == test_data['list']
    assert not hasattr(conf, '_private')

    conf = config.Config.get_config(
        files=[config_file, 'non_existent.yaml'], create=False
    )
    assert conf.name == test_data['name']

    new_config_file = os.path.join(temp_dir, 'new_folder', 'new_config.yaml')
    conf = config.Config.get_config(
        files=[new_config_file], create=True, initial_value='test'
    )
    assert os.path.exists(new_config_file)
    assert conf.initial_value == 'test'
    assert conf.get('non_existent', 'default') == 'default'
    assert str(conf) == repr(conf)
    assert isinstance(str(conf), str)

    assert 'new_key' not in conf
    conf['new_key'] = 'new_value'
    assert conf['new_key'] == 'new_value'


def test_config_query(temp_dir):
    config_file = os.path.join(temp_dir, 'config.yaml')
    test_data = {
        'parent': {
            'child_1': {
                'name': 'Rob',
            },
            'name': 'Parent',
        }
    }
    conf = config.Config(config_file=config_file, create=False, **test_data)
    assert conf.query_get('parent.child_1.name') == 'Rob'
    assert conf.query_get('parent.child_2.name', 'Not Exist') == 'Not Exist'
    conf.query_set('parent.child_1.name', 'Changed')
    assert conf.query_get('parent.child_1.name') == 'Changed'
    with pytest.raises(KeyError):
        conf.query_set('parent.child_2.name', 'Changed')
    conf.query_del('parent.child_1.name')
    assert conf.query_get('parent.child_1.name', 'Deleted') == 'Deleted'
    with pytest.raises(KeyError):
        conf.query_del('parent.child_2.name')


def test_config_del(temp_dir):
    config_file = os.path.join(temp_dir, 'config.yaml')
    test_data = {
        'key_1': 1,
        'key_2': 2,
    }
    conf = config.Config(config_file=config_file, create=False, **test_data)
    del conf['key_1']
    assert 'key_1' not in conf
    with pytest.raises(KeyError):
        del conf['key_1']



import time

from clifire import application, command, out


class OutCommand(command.Command):
    _name = 'out'

    def fire(self):
        out.info('This is an informational message')
        out.success('Operation completed successfully')
        out.warn('Warning: check your configuration')
        out.error('Critical error occurred')

        print('')
        print('Debug')
        print('-' * 80)
        out.debug('Debug: variable x = 42')
        sample_dict = {'key': 'value', 'numbers': [1, 2, 3]}
        out.var_dump(sample_dict)

        print('')
        print('Tables')
        print('-' * 80)
        data = [
            {'name': 'Luke', 'age': 18, 'is_student': True},
            {'name': 'Elizabeth', 'age': 101, 'is_student': False},
        ]
        out.table(data, border=True, title='Contacts')

        print('')
        print('Live text')
        print('-' * 80)
        live_text = out.LiveText('Starting...')
        time.sleep(1)
        live_text.info('Process running')
        time.sleep(1)
        live_text.warn('Retrying operation')
        time.sleep(1)
        live_text.success('Operation completed', end=False)
        time.sleep(1)


def main():
    app = application.App()
    app.add_command(OutCommand)
    app.fire('out -v')


if __name__ == '__main__':
    main()



import os
from typing import Any, Dict, List

import yaml
from clifire import out


class Config:
    @classmethod
    def get_config(cls, files: List[str], create: bool = False, **kwargs):
        def path(*args) -> str:
            expand_path = os.path.join(
                *(a.replace('~', os.path.expanduser('~')) for a in args)
            )
            return os.path.abspath(expand_path)

        if not files:
            return None
        for file in files:
            config_file = path(file)
            out.debug(f'Try read config file {config_file}')
            if os.path.exists(config_file):
                out.debug2('Exists, reading...')
                config = Config(config_file=config_file, **kwargs)
                config.read()
                return config
            out.debug2('Not exist')
        config = Config(config_file=path(files[0]), **kwargs)
        if create:
            config.write()
        return config

    def __init__(self, config_file, **kwargs):
        self._config_file = config_file
        self._config_path = os.path.dirname(config_file)
        self.__dict__.update(kwargs)

    def query_get(self, query, default=None):
        item = self.__dict__
        for key in query.split('.'):
            if key not in item:
                return default
            item = item[key]
        return item

    def query_set(self, query, value):
        item = self.__dict__
        keys = query.split('.')
        while keys:
            key = keys.pop(0)
            if not keys:
                item[key] = value
                return
            if key not in item:
                raise KeyError(f'{query} not exists')
            item = item[key]

    def query_del(self, query):
        item = self.__dict__
        keys = query.split('.')
        while keys:
            key = keys.pop(0)
            if not keys:
                del item[key]
                return
            if key not in item:
                raise KeyError(f'{query} not exists')
            item = item[key]

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __delitem__(self, key):
        if key in self.__dict__:
            del self.__dict__[key]
        else:
            raise KeyError(f'Key "{key}" not found')

    def __contains__(self, key):
        return key in self.__dict__

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return self.__str__()

    def _safe_dict(self, data_dict: Dict[str, Any]):
        return {k: v for k, v in data_dict.items() if not k.startswith('_')}

    def read(self):
        if not os.path.exists(self._config_file):
            return False
        with open(self._config_file, encoding='utf-8') as file:
            self.__dict__.update(self._safe_dict(yaml.safe_load(file) or {}))
        return True

    def write(self):
        out.debug(f'Write config file {self._config_file}')
        config_dir = os.path.dirname(self._config_file)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        with open(self._config_file, 'w', encoding='utf-8') as file:
            yaml.safe_dump(self._safe_dict(self.__dict__), file)



import time
from io import StringIO

import pytest
from clifire import application, out
from rich.console import Console


@pytest.fixture(autouse=True)
def reset_current_app():
    application.App.current_app = None
    yield


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_info(capsys):
    out.info('Info sample message')
    assert 'Info sample message\n' == output(capsys)


def test_success(capsys):
    out.success('Success sample message', icon=False)
    assert 'Success sample message\n' == output(capsys)


def test_warn(capsys):
    out.warn('Warn sample message', icon=False)
    assert 'Warn sample message\n' == output(capsys)


def test_error(capsys):
    out.warn('Error sample message', icon=False)
    assert 'Error sample message\n' == output(capsys)


def test_critical(capsys):
    with pytest.raises(SystemExit) as excinfo:
        out.critical('Critical sample message 1')
    assert '1' == str(excinfo.value)
    assert 'Critical sample message 1\n' == output(capsys)
    with pytest.raises(SystemExit) as excinfo:
        out.critical('Critical sample message', 99)
    assert '99' == str(excinfo.value)
    assert 'Critical sample message\n' == output(capsys)


def test_debug(capsys):
    out.debug('Debug sample message')
    assert 'Debug sample message\n' == output(capsys)


def test_live_text():
    buffer = StringIO()
    original_console = out.CONSOLE
    try:
        out.CONSOLE = Console(
            file=buffer, force_terminal=True, color_system='truecolor'
        )
        live = out.LiveText('First message')
        time.sleep(0.5)
        assert 'First message' in out.ansi_clean(buffer.getvalue())
        live.info('Info message')
        time.sleep(0.5)
        assert 'Info message' in out.ansi_clean(buffer.getvalue())
        live.warn('Warn message')
        time.sleep(0.5)
        assert 'Warn message' in out.ansi_clean(buffer.getvalue())
        live.error('Error message', end=False)
        time.sleep(0.5)
        assert 'Error message' in out.ansi_clean(buffer.getvalue())
        live.success('Success message', end=False)
        time.sleep(0.5)
        assert 'Success message' in out.ansi_clean(buffer.getvalue())
    finally:
        out.CONSOLE = original_console
    assert live._running is True
    live.cancel()
    assert live._running is False


def test_live_concurrent():
    buffer = StringIO()
    original_console = out.CONSOLE
    try:
        out.CONSOLE = Console(
            file=buffer, force_terminal=True, color_system='truecolor'
        )
        live_1 = out.live('First message')
        live_2 = out.live('Second message')
        assert live_1 == live_2
    finally:
        out.CONSOLE = original_console


def test_live_text_threading():
    buffer = StringIO()
    original_console = out.CONSOLE
    try:
        out.CONSOLE = Console(
            file=buffer, force_terminal=True, color_system='truecolor'
        )

        live = out.LiveText('First message')
        assert live.is_alive is True
        live.stop()
        assert live.is_alive is False
        live.start()
        assert live.is_alive is True
        live.start()
        assert live.is_alive is True
        live.stop()
        assert live.is_alive is False

        live = out.LiveText('First message')
        live.error('Error message')
        assert 'Error message' in out.ansi_clean(buffer.getvalue())
        assert live.is_alive is False

        live = out.LiveText('First message')
        assert live.is_alive is True
        live.success('Success message')
        assert live.is_alive is False
        assert 'Success message' in out.ansi_clean(buffer.getvalue())

        live = out.LiveText('First message')
        live.warn('Warn message', end=True)
        assert live.is_alive is False
        assert 'Warn message' in out.ansi_clean(buffer.getvalue())

        live = out.LiveText('First message')
        live.info('Info message', end=True)
        assert live.is_alive is False
        assert 'Info message' in out.ansi_clean(buffer.getvalue())
    finally:
        out.CONSOLE = original_console


def test_table(capsys):
    data = [
        {'name': 'Luke', 'age': 18, 'is_student': True},
        {'name': 'Elizabeth', 'age': 101, 'is_student': False},
    ]
    buffer = StringIO()
    original_console = out.CONSOLE
    try:
        out.CONSOLE = Console(
            file=buffer, force_terminal=True, color_system='truecolor'
        )
        out.table(data, border=False, title='My title', style_cols='red')
        raw_output = buffer.getvalue()
        assert '\x1b[' in raw_output
        assert '\x1b[31m' in raw_output or '\x1b[91m' in raw_output
    finally:
        out.CONSOLE = original_console

    out.table(data, border=False, title='My title')
    printed = output(capsys)
    assert 'My title' in printed
    assert 'Is student' in printed
    assert '│' not in printed
    assert 'Elizabeth' in printed

    out.table(data, border=True, title='My title')
    printed = output(capsys)
    assert 'My title' in printed
    assert 'Is student' in printed
    assert '│' in printed
    assert '│ Elizabeth │' in printed

    out.table(data, border=True, padding=(0, 0))
    printed = output(capsys)
    assert '│Elizabeth│' in printed

    out.table([])
    printed = output(capsys)
    assert '' == printed


def test_var_dump(capsys):
    out.var_dump([1, 2, 3, 4])
    assert '[1, 2, 3, 4]' in output(capsys)

    out.var_dump(
        {
            'number': 1,
            'list': ['a', 'b', 'c'],
            'dict': {
                'key a': 'value a',
                'key b': 'value b',
            },
        }
    )
    printed = output(capsys)
    assert "'number': 1," in printed
    assert "'list': ['a', 'b', 'c']," in printed
    assert "'dict': {'key a': 'value a'," in printed


def test_ask(capsys, monkeypatch):
    monkeypatch.setattr('builtins.input', lambda: 'y')
    response = out.ask('You are sure?')
    assert response == 'y'
    printed = output(capsys)
    assert 'You are sure?' in printed
    assert '[y/n]' in printed

    monkeypatch.setattr('builtins.input', lambda: 'answer')
    response = out.ask('What is your name?', choices=None)
    assert response == 'answer'
    printed = output(capsys)
    assert 'What is your name?' in printed
    assert '[y/n]' not in printed

    monkeypatch.setattr('builtins.input', lambda: 'yes')
    response = out.ask('Continue?', choices=['yes', 'no'])
    assert response == 'yes'
    printed = output(capsys)
    assert 'Continue?' in printed
    assert '[yes/no]' in printed


def test_rule(capsys):
    out.rule('My rule')
    printed = output(capsys)
    assert 'My rule' in printed
    assert '─' in printed

    out.rule('My rule')
    printed = output(capsys)
    assert 'My rule' in printed
    assert '─' in printed

    live = out.live('Message in live')
    assert 'Message in live' in live._text
    out.rule('My rule')
    assert 'My rule' in live._text
    printed = output(capsys)
    assert 'My rule' in printed
    live.cancel()
    printed = output(capsys)
    assert 'My rule' not in printed


def test_no_ansi(capsys):
    out.setup(ansi=False, show_icons=False)
    assert out.CONSOLE.no_color is True
    out.success('Success')
    assert '✓ Success' in output(capsys)
    out.warn('Warn')
    assert '▲ Warn' in output(capsys)
    out.error('Error')
    assert '✗ Error' in output(capsys)

    live = out.live('Message in live')
    assert 'Message in live' in live._text
    live.success('Success', end=False)
    time.sleep(1)
    assert '✓ Success' in live._text
    live.error('Error', end=False)
    assert '✗ Error' in live._text
    live.cancel()

    output(capsys)
    out.setup(ansi=True, show_icons=False)
    assert out.CONSOLE.no_color is False
    out.success('Success')
    assert '✓' not in output(capsys)
    out.warn('Warn')
    assert '▲' not in output(capsys)
    out.error('Error')
    assert '✗' not in output(capsys)
    out.success('Success', icon=True)
    assert '✓ Success' in output(capsys)
    out.warn('Warn', icon=True)
    assert '▲ Warn' in output(capsys)
    out.error('Error', icon=True)
    assert '✗ Error' in output(capsys)



import inspect
import re
import shlex
from typing import List, Optional, Type, Union

from clifire import out


def get_current_app():
    from clifire import application

    return application.App.current_app


def fire(func):
    def wrapper(*args, **kwargs):
        cmd = args[0]
        signature = inspect.signature(func)
        for arg in list(signature.parameters.keys())[1:]:
            var = arg[1:] if arg.startswith('_') else arg
            if hasattr(cmd, var):
                kwargs[arg] = getattr(cmd, var)
        return func(*args, **kwargs)

    command_name = getattr(func, '_command_name', func.__name__)
    command_name = re.sub(r'([A-Z_])', '.', command_name).lower()
    class_name = ''.join(
        word.capitalize() for word in command_name.split('.') if word
    )
    doc = (func.__doc__ or '').strip()
    attrs = {
        '_name': command_name,
        '_help': doc.splitlines()[0] if doc.splitlines() else doc,
        'fire': wrapper,
    }
    signature = inspect.signature(func)
    pos = 0
    helps = {k: [] for k in list(signature.parameters.keys())[1:]}
    doc = [d.strip() for d in doc.splitlines() if d]
    current_help = False
    while doc:
        line = doc.pop(0)
        parts = line.split(':')
        if parts[0] in helps:
            current_help = parts[0]
            line = parts[1]
        if not current_help:
            continue
        helps[current_help].append(line.strip())
    for name, param in list(signature.parameters.items())[1:]:
        pos += 1
        var_name = name[1:] if name.startswith('_') else name
        attrs[var_name] = Field(
            pos=None if name.startswith('_') else pos,
            help=' '.join(helps.get(name, '')),
            default=param._default,
            force_type=param._annotation,
            alias=[var_name[0]],
        )
    class_command = type(class_name, (Command,), attrs)
    app = get_current_app()
    app.add_command(class_command)
    return wrapper


class Field:
    def __init__(
        self,
        pos: int = False,
        help: str = '',
        default: str = None,
        alias: Union[str, List[str], None] = None,
        force_type: Optional[Type] = None,
    ):
        self.name = 'unknow'
        self.pos = pos
        self.help = help
        alias = [] if alias is None else alias
        self.alias = [alias] if isinstance(alias, str) else alias
        self.default = default
        self.value = default
        self.is_option = bool(pos is False or pos is None)
        self.is_required = default is None
        if force_type is None:
            if self.default is not None:
                force_type = type(self.default)
            elif self.is_option:
                force_type = bool
            else:
                force_type = str
        self.type = force_type

    def convert(self, value):
        try:
            if self.type == list:
                return value.split(',')
            elif self.type == bool and value is None:
                return True if self.default is None else not self.default
            return self.type(value)
        except Exception as exc:
            msg = f'with he value "{value}" must be {self.type.__name__}'
            raise FieldException(self, msg) from exc


class FieldException(Exception):
    def __init__(self, field: Field, msg: str):
        self.field = field
        field_type = 'option' if self.field.is_option else 'argument'
        super().__init__(f'The {field_type} "{field.name}" {msg}')


class Command:
    _name = ''
    _help = None

    def __init__(self, app, command_line: str = ''):
        self._fields = {}
        self._argument_names = []
        self._options = {}
        self.app = app
        self._fields_update()
        self.command_line = command_line
        self.extra_args = []
        self.init()

    @property
    def context(self):
        return self.app.context

    def _fields_update(self):
        for name in dir(self):
            field = getattr(self, name)
            if isinstance(field, Field):
                field.name = name
                self._fields[name] = field
        self._argument_names = sorted(
            [key for key, val in self._fields.items() if not val.is_option],
            key=lambda k: self._fields[k].pos,
        )
        self._options = {}
        for name, field in self._fields.items():
            if not field.is_option:
                continue
            self._options[name] = field
            for alias in field.alias:
                if alias.startswith('-'):
                    alias = alias[2:] if alias.startswith('--') else alias[1:]
                alias = alias.replace('-', '_')
                if alias in self._options:
                    raise CommandException(f'Duplicate option alias "{alias}"')
                self._options[alias] = name

    def _fields_check(self):
        for name, field in self._fields.items():
            value = getattr(self, name)
            if isinstance(value, Field):
                if field.is_required:
                    raise FieldException(field, 'is required')
                setattr(self, name, field.default)

    def _parse_command_line(self, command_line: str):
        out.debug(f'Parse command line: {command_line}')
        tokens = shlex.split(command_line)
        command_parts = self._name.split('.')
        argument_index = 0
        index = 0
        while index < len(tokens):
            token = tokens[index]
            if token in command_parts:
                command_parts.remove(token)
                index += 1
            elif token.startswith('-'):
                consumed = self._handle_option(tokens, index)
                index += consumed
            else:
                consumed = self._handle_argument(tokens, index, argument_index)
                if consumed > 0:
                    argument_index += 1
                    index += consumed
                else:
                    self.extra_args.append(token)
                    index += 1

    def _handle_option(self, tokens: List[str], index: int) -> int:
        token = tokens[index]
        option_str = token[2:] if token.startswith('--') else token[1:]
        name, value = (
            option_str.split('=', 1)
            if '=' in option_str
            else (option_str, None)
        )
        name = name.replace('-', '_')
        option_field = self._find_option(name)
        if not option_field:
            self.extra_args.append(token)
            return 1
        consumed = 1
        if value is None and option_field['field'].type != bool:
            next_token = tokens[index + 1] if index + 1 < len(tokens) else None
            if not next_token.startswith('-'):
                value = next_token
                consumed = 2
        parsed_value = option_field['field'].convert(value)
        option_name = option_field['name']
        if option_field['is_global']:
            self.app.set_option(option_name, parsed_value)
            out.debug2(f'Global option "{option_name}" = {parsed_value}')
        else:
            setattr(self, option_name, parsed_value)
            out.debug2(f'Option "{option_name}" = {parsed_value}')
        return consumed

    def _find_option(self, name: str) -> dict:
        def _get_options(name: str) -> dict:
            if name in self.app.options:
                return self.app.options
            if name in self._options:
                return self._options
            return {}

        options = _get_options(name)
        if not options:
            return None
        field_name = options[name] if isinstance(options[name], str) else name
        return {
            'field': options[field_name],
            'name': field_name,
            'is_global': name in self.app.options,
        }

    def _handle_argument(
        self, tokens: List[str], index: int, argument_index: int
    ) -> int:
        if argument_index >= len(self._argument_names):
            return 0
        field_name = self._argument_names[argument_index]
        field = self._fields[field_name]
        if field.type == list:
            list_values = []
            consumed = 0
            for i in range(index, len(tokens)):
                token = tokens[i]
                list_values.append(token)
                consumed += 1
            setattr(self, field_name, list_values)
            out.debug2(f'Argument "{field_name}" = {list_values}')
            return consumed
        token = tokens[index]
        value = field.convert(token)
        setattr(self, field_name, value)
        out.debug2(f'Argument "{field_name}" = {value}')
        return 1

    def parse(self, command_line: str):
        self._parse_command_line(command_line)
        self._fields_check()

    def launch(self, command_line: str):
        self.command_line = command_line
        out.debug(f'Launching command "{self._name}"')
        self.parse(command_line)
        out.debug(f'Running command "{self._name}"')
        return self.fire()

    def init(self):
        pass

    def fire(self):
        raise NotImplementedError


class CommandException(Exception):
    pass



from clifire import command, out


@command.fire
def greet(cmd, name: str = 'World', _end_char: str = '?'):
    '''
    Greets the user.

    Args:
        name: Name of the user to greet. Defaults to "World".
        _end_char: Character to use at the end of the greeting. Defaults to "?".
    '''
    result = cmd.app.shell('whoami')
    out.info(f'System user: {result.stdout}')
    out.success(f'Hello {name}{_end_char}')



from clifire import command, out


@command.fire
def hello(cmd, user: str = '', _sudo: bool = False):
    '''
    Display a greeting on the console

    Args:
        user: Name of the user to greet. If empty, it will use the current
            system user.
        _sudo: It will run the command with sudo privileges.
    '''
    out.debug(f'Sudo = {_sudo}')
    if not user:
        sudo = 'sudo' if _sudo else ''
        user = cmd.app.shell(f'{sudo} whoami').stdout
    out.info(f'Hi {user}!')


@command.fire
def nodoc(cmd):
    out.info('Command without doc')


@command.fire
def ab(cmd):
    '''
    doc_ab
    '''
    out.info('def_ab')


@command.fire
def ab_cd(cmd):
    '''
    doc_ab_cd
    '''
    out.info('def_ab_cd')


@command.fire
def ab_ef_gh(cmd):
    '''
    doc_ab_ef_gh
    '''
    out.info('def_ab_ef_gh')


@command.fire
def zz_command(cmd):
    '''
    doc_zz_command
    '''
    out.info('def_zz_command')



from clifire import application, command, out


class ConfigCommand(command.Command):
    _name = 'config'
    _help = 'Show config vars'

    def fire(self):
        self.app.config.my_new_var = 'My new var'
        out.var_dump(self.app.config)


def main():
    app = application.App(config_files=['~/.myapp.yml'], config_create=True)
    app.add_command(ConfigCommand)
    app.fire()


if __name__ == '__main__':
    main()



import atexit
import re
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from rich import traceback
from rich.console import Console
from rich.live import Live
from rich.prompt import Prompt
from rich.table import Table

traceback.install(
    show_locals=True,
    theme='monokai',
    suppress=[],
    max_frames=2,
)

CONSOLE = Console()
CONSOLE_WIDTH = CONSOLE.width

COLOR_NORMAL = 'white'
COLOR_DEBUG = 'grey50'
COLOR_DEBUG2 = 'bright_black'
COLOR_INFO = 'blue'
COLOR_SUCCESS = 'green'
COLOR_WARN = 'yellow'
COLOR_ERROR = 'red'

ICON_SHOW = False
ICON_SUCCESS = '✓'
ICON_ERROR = '✗'
ICON_WARN = '▲'


def setup(ansi: bool = True, show_icons: bool = None) -> None:
    global ICON_SHOW, CONSOLE, CONSOLE_WIDTH
    if show_icons is not None:
        ICON_SHOW = show_icons
    if ansi:
        CONSOLE.no_color = False
        CONSOLE.highlight = True
        CONSOLE.soft_wrap = True
        CONSOLE.width = CONSOLE_WIDTH
        debug('Console output setup with ANSI')
    else:
        CONSOLE.no_color = True
        CONSOLE.highlight = False
        CONSOLE.soft_wrap = False
        CONSOLE.width = 10000
        debug('Console output setup with no ANSI')


class LiveText:
    def __init__(self, text: str = '', refresh_per_second: int = 10):
        self._live = None
        self._running = False
        self._thread = None
        self._text = ''
        atexit.register(self.cancel)
        self.elapsed_time = 0
        self.refresh_per_second = refresh_per_second
        self.info(text)
        self.start()

    @property
    def is_alive(self):
        return self._running

    def _start(self):
        with Live(
            console=CONSOLE,
            refresh_per_second=self.refresh_per_second,
        ) as live:
            self._live = live
            self._live.update(self._text)
            start_time = time.time()
            while self._running:
                self.elapsed_time = time.time() - start_time
                display_message = (
                    f'[grey0]({self.elapsed_time: .1f}s)[/grey0] '
                    f'{self._text}\n'
                )
                self._live.update(display_message)
                time.sleep(1 / self.refresh_per_second)
            self._live.update(self._text)

    def start(self):
        self._running = True
        if self._thread and self._thread.is_alive():
            return
        self._thread = threading.Thread(target=self._start, daemon=True)
        self._thread.start()

    def cancel(self):
        self._text = ''
        self.stop()

    def stop(self):
        if self._text != '':
            CONSOLE.print(self._text)
        self._running = False
        if self._thread and self._thread.is_alive():
            if self._live:
                self._live.transient = True
            self._text = text_color('')
            self._thread.join(timeout=1.0)
            if self._live:
                self._live.update(self._text)
                self._live.stop()

    def info(self, text: str, end=False):
        self._text = text_color(text, color=COLOR_INFO)
        if end:
            self.stop()

    def warn(self, text: str, end=False, icon: bool = None):
        self._text = text_color(
            text, color=COLOR_WARN, icon=ICON_WARN, show_icon=icon
        )
        if end:
            self.stop()

    def success(self, text: str, end=True, icon: bool = None):
        self._text = text_color(
            text, color=COLOR_SUCCESS, icon=ICON_SUCCESS, show_icon=icon
        )
        if end:
            self.stop()

    def error(self, text: str, end=True, icon: bool = None):
        self._text = text_color(
            text, color=COLOR_ERROR, icon=ICON_ERROR, show_icon=icon
        )
        if end:
            self.stop()


_current_live = None


def live(text: str = '', refresh_per_second: int = 10):
    global _current_live
    if _current_live:
        _current_live.info(text)
        _current_live.refresh_per_second = refresh_per_second
        if not _current_live.is_alive:
            _current_live.start()
        return _current_live
    _current_live = LiveText(text, refresh_per_second=refresh_per_second)
    return _current_live


def table(
    data: List[Dict[str, Any]],
    title: str = '',
    border: bool = True,
    show_header: bool = True,
    style_cols: Optional[Union[Dict[str, str], str]] = None,
    padding: Optional[Tuple] = None,
    style: Optional[str] = None,
):
    if not data:
        return
    keys = list(data[0].keys())
    tbl = Table(title=title, show_header=show_header, style=style)
    if not border:
        tbl.box = None
        tbl.padding = (0, 2)
    if padding is not None:
        tbl.padding = padding
    for key in keys:
        justify = 'right' if isinstance(data[0][key], (int, float)) else 'left'
        _style_col = None
        if isinstance(style_cols, dict):
            _style_col = style_cols.get(key, None)
        elif isinstance(style_cols, str):
            _style_col = style_cols
        tbl.add_column(
            key.replace('_', ' ').capitalize(),
            justify=justify,
            style=_style_col,
        )
    for row in data:
        tbl.add_row(*(str(row.get(key, '')) for key in keys))
    CONSOLE.print(tbl)


def ansi_clean(text: str) -> str:
    return re.sub(r'\x1B\[[0-?]*[ -/]*[@-~]', '', text)


def text_color(
    text: str,
    color: str = COLOR_NORMAL,
    icon: str = None,
    show_icon: bool = False,
) -> str:
    if show_icon is None:
        show_icon = ICON_SHOW
    text = str(text)
    if icon and not text.startswith(icon):
        if CONSOLE.no_color is True or show_icon:
            text = f'{icon} {text}'
    return f'[{color}]{text}[/{color}]'


def _print(
    text: str,
    color: str = COLOR_NORMAL,
    icon: str = None,
    show_icon: bool = None,
) -> None:
    text = text_color(text, color=color, icon=icon, show_icon=show_icon)
    CONSOLE.print(text)


def info(text: str) -> None:
    _print(text, COLOR_INFO)


def success(text: str, icon: bool = None) -> None:
    _print(text, color=COLOR_SUCCESS, icon=ICON_SUCCESS, show_icon=icon)


def warn(text: str, icon: bool = None) -> None:
    _print(text, color=COLOR_WARN, icon=ICON_WARN, show_icon=icon)


def error(text: str, icon: bool = None) -> None:
    _print(text, color=COLOR_ERROR, icon=ICON_ERROR, show_icon=icon)


def critical(text: str, code: int = 1) -> None:
    error(text, icon=False)
    sys.exit(code)


def _debug(text: str, color: str) -> None:
    from clifire import application

    app = application.App.current_app
    if not app or app.get_option('verbose'):
        _print(text, color)


def debug(text: str) -> None:
    _debug(text, COLOR_DEBUG)


def debug2(text: str) -> None:
    _debug(f'· {text}', COLOR_DEBUG2)


def var_dump(var) -> None:
    CONSOLE.print(var, highlight=True)


def ask(text: str, choices: List[str] = False):
    if choices is False:
        choices = ['y', 'n']
    return Prompt.ask(
        text, choices=choices, default=choices[0] if choices else None
    )


def rule(text: str) -> None:
    global _current_live
    if _current_live:
        _current_live.info(text)
    CONSOLE.rule(f'[bold blue]{text}', align='left', style='blue')



import runpy
import sys

from clifire import out

from tests.sample.out import app


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_main_block(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', ['tests/sample/out.py', 'help', 'info'])
    if 'tests.sample.out' in sys.modules:
        del sys.modules['tests.sample.out']
    for key in list(sys.modules.keys()):
        if key.startswith('tests.sample.'):
            del sys.modules[key]
    runpy.run_module('tests.sample.out', run_name='__main__')
    assert 'Show a message text with info style.' in output(capsys)


def test_sample_out_info(capsys):
    app.fire('help info')
    printed = output(capsys)
    assert 'Show a message text with info style.' in printed
    assert 'Message text to show ' in printed
    assert 'A sample name ' in printed
    assert ' -t ' in printed
    assert '--color' in printed

    app.fire('info "Hello World"')
    printed = output(capsys)
    assert 'Hello World' in printed

    app.fire('info "Hello World" --type warn')
    printed = output(capsys)
    assert 'Hello World' in printed

    app.fire('info "Hello World" --type error')
    printed = output(capsys)
    assert 'Hello World' in printed

    app.fire('info "Hello World" --type not-exists')
    printed = output(capsys)
    assert 'Type "not-exists" unknow' in printed

    app.fire('info "Hello World" Elvis --type warn')
    printed = output(capsys)
    assert 'Elvis: Hello World' in printed


def test_sample_out_info_ask(capsys):
    app.fire('help info ask')
    printed = output(capsys)
    assert 'Show a message text like a question.' not in printed
    assert 'Show a message how a question.' in printed
    assert ' -t ' in printed
    assert '--color' in printed
    assert '--char' in printed

    app.fire('info ask "Hello World"')
    printed = output(capsys)
    assert 'Hello World?' in printed

    app.fire('info ask "Hello World" --type warn --char !')
    printed = output(capsys)
    assert 'Hello World!' in printed

    app.fire('info ask "Hello World" --type error --end-char !')
    printed = output(capsys)
    assert 'Hello World!' in printed

    app.fire('info ask "Hello World" --type not-exists --end-char !')
    printed = output(capsys)
    assert 'Type "not-exists" unknow' in printed


def test_sample_out_live(capsys):
    app.fire('live 2')
    printed = output(capsys)
    assert 'Regresive count for 2 ended!' in printed



from clifire import command, out


class CommandHelp(command.Command):
    _name = 'help'
    _help = 'Show this help'

    command = command.Field(
        pos=1,
        help='Command to show help',
        default=['help'],
    )

    def print(self, txt):
        out._print(txt)

    def get_help(self, cmd) -> str:
        lines = (cmd._help or cmd.__doc__ or '').strip().splitlines()
        return lines[0] if lines else ''

    def print_description(self, cmd):
        title = 'Description:'
        self.print(f'[bold]{title}[/bold]')
        self.print(f'  {self.get_help(cmd)}')
        self.print('')

    def print_usage(self, cmd):
        title = 'Usage:'
        self.print(f'[bold]{title}[/bold]')
        args = [
            f'<{n}>' if cmd._fields[n].is_required else f'[<{n}>]'
            for n in cmd._argument_names
        ]
        cmd_name = cmd._name.replace('.', ' ')
        args_str = ' '.join(args)
        self.print(f'  [cyan]{cmd_name}[/cyan] \\[options] {args_str}')
        self.print('')

    def print_arguments(self, cmd):
        data = []
        for name in cmd._argument_names:
            field = cmd._fields[name]
            data.append({'name': name, 'help': field.help})
        if not data:
            return False
        title = 'Arguments:'
        self.print(f'[bold]{title}[/bold]')
        out.table(
            data, border=False, show_header=False, style_cols={'name': 'cyan'}
        )
        self.print('')

    def print_options(self, cmd):
        options = {}
        for name in cmd._options:
            field = cmd._options[name]
            if isinstance(field, str):
                field = cmd._fields[field]
            name = name.replace('_', '-')
            name = f'-{name}' if len(name) == 1 else f'--{name}'
            options.setdefault(field, []).append(name)
        if not options:
            return False
        data = []
        for field, names in options.items():
            short = ', '.join(sorted([n for n in names if len(n) == 2]))
            name = ', '.join(sorted([n for n in names if len(n) != 2]))
            data.append({'short': short, 'name': name, 'help': field.help})
        title = 'Options:'
        self.print(f'[bold]{title}[/bold]')
        out.table(
            sorted(data, key=lambda d: d['short']),
            border=False,
            show_header=False,
            style_cols={'short': 'cyan', 'name': 'cyan'},
        )
        self.print('')

    def print_options_global(self):
        options = {}
        for name in self.app.options:
            field = self.app.options[name]
            if isinstance(field, str):
                field = self.app.options[field]
            name = name.replace('_', '-')
            name = f'-{name}' if len(name) == 1 else f'--{name}'
            options.setdefault(field, []).append(name)
        if not options:
            return False
        data = []
        for field, names in options.items():
            short = ', '.join(sorted([n for n in names if len(n) == 2]))
            name = ', '.join(sorted([n for n in names if len(n) != 2]))
            data.append({'short': short, 'name': name, 'help': field.help})
        title = 'Global options:'
        self.print(f'[bold]{title}[/bold]')
        out.table(
            sorted(data, key=lambda d: d['short']),
            border=False,
            show_header=False,
            style_cols={'short': 'cyan', 'name': 'cyan'},
        )
        self.print('')

    def print_commands(self, cmd):
        groups = {}
        group_name = (
            '' if self.command == ['help'] else '.'.join(self.command + [''])
        )
        for cls in self.app.commands.values():
            name = cls._name
            if group_name:
                if not name.startswith(group_name):
                    continue
                name = name[len(group_name) :]
            parts = name.split('.')
            group = ' '.join(parts[:-1])
            groups.setdefault(group, []).append(
                {
                    'name': name.replace('.', ' '),
                    'help': self.get_help(cls),
                }
            )
        if not groups:
            return
        title = 'Available Commands:'
        self.print(f'[bold]{title}[/bold]')
        data = []
        for key in sorted(groups.keys()):
            if key and key != cmd._name:
                data.append({'name': f'\n[green]{key}[/green]', 'help': ''})
            data += sorted(groups[key], key=lambda item: item['name'])
        out.table(
            data,
            border=False,
            show_header=False,
            style_cols={'name': 'cyan'},
            padding=(0, 2) if key else None,
        )
        self.print('')

    def fire(self):
        cmd = self.app.get_command(' '.join(self.command))
        self.print_description(cmd)
        self.print_usage(cmd)
        self.print_arguments(cmd)
        self.print_options(cmd)
        self.print_options_global()
        self.print_commands(cmd)



import os

from clifire import command, out


@command.fire
def doc_build(cmd):
    '''
    Build the docs
    '''
    bash = 'rye run mkdocs build'
    cmd.app.shell(bash, path='./docs', capture_output=False)


@command.fire
def doc_serve(cmd):
    '''
    Serve the docs
    '''
    bash = 'rye run mkdocs serve'
    cmd.app.shell(bash, path='./docs', capture_output=False)


@command.fire
def doc_record(cmd, filename: str, _force: bool = False):
    '''
    Record terminal session

    Args:
        filename: name of the file with which you want to record the session
        -f, --force: remove files if already exists
    '''
    result = cmd.app.shell('which asciinema')
    if not result.stdout:
        out.warn('You can install with `apt install asciinema`')
        out.critical('`asciinema` not found.')
    result = cmd.app.shell('which svg-term')
    if not result.stdout:
        out.warn('You can install with `npm install -g svg-term-cli`')
        out.critical('`svg-term` not found.', code=2)

    file_cast = cmd.app.path(f'./docs/docs/assets/records/{filename}.cast')
    if os.path.exists(file_cast) and _force is True:
        os.remove(file_cast)
    if os.path.exists(file_cast):
        msg = 'File cast already exists, you can use --force for overwrite'
        out.critical(msg, code=3)
    out.info(f'Record file in {file_cast}')

    file_svg = cmd.app.path(f'./docs/docs/assets/records/{filename}.svg')
    if os.path.exists(file_svg) and _force is True:
        os.remove(file_svg)
    if os.path.exists(file_svg):
        msg = 'File SVG already exists, you can use --force for overwrite'
        out.critical(msg, code=4)

    args = ''
    shell = os.environ.get('SHELL')
    if shell == '/bin/fish':
        args = '-c \'fish --private -C "function fish_prompt; echo -n \\" > \\"; end"\''
    bash = f'asciinema rec {file_cast} {args}'
    pwd = cmd.app.path('./docs/docs')
    out.info(f'You are now in {pwd}')
    cmd.app.shell(bash, path=pwd, shell=True, check=True, capture_output=False)

    with open(file_cast) as f:
        lines = f.readlines()
    lines = [ln for ln in lines if 'fish is running in private mode' not in ln]
    with open(file_cast, 'w') as f:
        f.writelines(lines)

    cmd.app.shell(
        f'svg-term --in {file_cast} --out {file_svg} --window --padding 10'
    )
    out.success(f'File SVG created in {file_svg}')
    out.info('You can use it in doc with:')
    out.info(f'![{filename.title()}](assets/records/{filename}.svg)')


@command.fire
def doc_publish(cmd):
    '''
    Publish docs in https://rlizana.github.io/clifire
    '''
    bash = 'mkdocs gh-deploy'
    cmd.app.shell(
        bash, path='./docs', shell=True, check=True, capture_output=False
    )



import time

from clifire import application, command, out


class CommandInfo(command.Command):
    '''
    Show a message text with info style.
    '''

    _name = 'info'

    text = command.Field(pos=1, help='Message text to show in the console')
    name = command.Field(pos=2, default='', help='A sample name')
    type = command.Field(
        help='Color can be: info, warn or error',
        default='info',
        alias=['t', 'color'],
    )

    def fire(self):
        if self.name:
            self.text = f'{self.name}: {self.text}'
        if self.type == 'info':
            out.info(self.text)
        elif self.type == 'warn':
            out.warn(self.text)
        elif self.type == 'error':
            out.error(self.text)
        else:
            out.warn(f'Type "{self.type}" unknow')


class CommandInfoAsk(command.Command):
    '''
    Show a message how a question.
    '''

    _name = 'info.ask'

    text = command.Field(
        pos=1,
        force_type=str,
        help='Message text to show in the console',
    )
    type = command.Field(
        help='Color can be: info, warn or error',
        default='info',
        alias=['t', 'color'],
    )
    end_char = command.Field(
        help='character for end the phrase',
        default='?',
        alias=['char'],
    )

    def fire(self):
        self.text = f'{self.text}{self.end_char}'
        if self.type == 'info':
            out.info(self.text)
        elif self.type == 'warn':
            out.warn(self.text)
        elif self.type == 'error':
            out.error(self.text)
        else:
            out.warn(f'Type "{self.type}" unknow')


class CommandLive(command.Command):
    _name = 'live'
    _help = 'Show a message text in live mode'

    time = command.Field(pos=1, force_type=int, help='Duration in seconds')

    def fire(self):
        time_sleep = 0.5
        live = out.LiveText('Starting ...')
        time.sleep(time_sleep)
        live.warn('Ready ...')
        time.sleep(time_sleep)
        live.success('Go ...', end=False)
        time.sleep(time_sleep)
        live.error('Trying again ...', end=False)
        time.sleep(time_sleep)
        live.success('Go ...', end=False)
        max_value = self.time
        value = 0
        while value < max_value:
            live.info(f'Count {value} to {max_value}')
            value += 1
            time.sleep(time_sleep)
        live.success(f'Regresive count for {max_value} ended!')


app = application.App(name='Out sample', version='1.0')
app.add_command(CommandInfo)
app.add_command(CommandInfoAsk)
app.add_command(CommandLive)

if __name__ == '__main__':
    app.fire()



from clifire import command


@command.fire
def test_legacy(cmd, _build: bool = False, python_version: str = '3.8.2'):
    '''
    Launch tests in docker image with specified Python version

    Args:
        python_version: Python version to use for the tests. (default: '3.8.2')
        _build: Force build the Docker image before running tests.
    '''

    image_name = f'clifire-py{python_version}'

    def docker_build():
        cmd.app.shell(
            f'docker build --build-arg PYTHON_VERSION={python_version} '
            f'-t {image_name} .',
            capture_output=False,
        )

    if _build:
        docker_build()
    elif image_name not in cmd.app.shell('docker images').stdout:
        docker_build()

    volumen_str = '-v ./src:/app/src -v ./tests:/app/tests'
    cmd.app.shell(
        f'docker run --rm {volumen_str} {image_name}',
        capture_output=False,
    )



from clifire import application, command, out


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_command_version(capsys):
    app = application.App(name='sample', version='1.0.99')
    app.fire('version')
    assert 'sample 1.0.99' in output(capsys)


def test_command_version_overwrite(capsys):
    class CommandVersion(command.Command):
        _name = 'version'

        def fire(self):
            out.info('Overwrite version message')

    app = application.App(
        name='sample', version='1.0.99', command_version=CommandVersion
    )
    app.fire('version')
    assert 'Overwrite version message' in output(capsys)



import os
import tempfile

import pytest
from clifire import application


@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield tmpdirname


def test_app_template(temp_dir):
    folder = os.path.join(os.path.dirname(__file__), 'sample', 'template')
    app = application.App(template_folder=folder)
    assert folder == app.template.path()
    kwargs = {
        'title': 'sample',
        'user': 'root',
        'items': ['test1', 'test2'],
    }
    content = app.template.render('sample.jinja2', **kwargs)
    assert 'sample' in content
    assert 'Hi root' in content
    assert 'test1' in content
    assert 'test2' in content

    temp_file = os.path.join(temp_dir, 'output.html')
    app.template.write('sample.jinja2', temp_file, **kwargs)
    with open(temp_file) as f:
        content = f.read()
    assert 'sample' in content
    assert 'Hi root' in content

    temp_file_mark = os.path.join(temp_dir, 'output_mark.html')
    mark = '<!-- MARK -->'
    content = app.template.write(
        'sample.jinja2', temp_file_mark, mark=mark, **kwargs
    )
    assert mark in content
    assert content.count(mark) == 2

    with open(temp_file_mark, 'w') as f:
        f.write(f'Header\n{mark}\nOld content\n{mark}\nFooter\n')
    content = app.template.write(
        'sample.jinja2', temp_file_mark, mark=mark, **kwargs
    )
    assert 'Header' in content
    assert 'Footer' in content
    assert 'Old content' not in content



import sys

import pytest
from clifire import application, command, out


class CommandContact(command.Command):
    _name = 'contact'

    name = command.Field(
        pos=1,
        help='Then contact name for show in console',
    )
    age = command.Field(
        pos=2,
        help='Then contact age',
        default=18,
    )
    option = command.Field(
        help='Sample option',
        default=False,
    )
    int_option = command.Field(
        help='Int option sample',
        default=1,
        alias=['o', 'int'],
    )
    str_option = command.Field(
        help='Str option sample',
        default='default',
        alias=['str'],
    )

    def init(self):
        self.option_init = command.Field(
            help='One dynamic option',
            default='dynamic',
        )

    def fire(self):
        out.info(f'Contact {self.name} with {self.age} years old')
        return True


@pytest.fixture(autouse=True)
def reset_current_app():
    application.App.current_app = None
    yield


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_command():
    app = application.App()
    cmd = command.Command(app)
    with pytest.raises(NotImplementedError):
        cmd.launch('')


def test_command_no_name():
    class CommandNoName(command.Command):
        pass

    app = application.App()
    with pytest.raises(command.CommandException):
        app.add_command(CommandNoName)


def test_command_exception(capsys):
    class CommandTest(command.Command):
        _name = 'test'

        def fire(self):
            raise command.CommandException('Test exception')

    app = application.App()
    app.add_command(CommandTest)
    with pytest.raises(SystemExit) as excinfo:
        app.fire('test')
    assert '30' == str(excinfo.value)
    assert 'Test exception' in output(capsys)


def test_command_not_exists(capsys):
    app = application.App()
    with pytest.raises(SystemExit) as excinfo:
        app.fire('NotExist')
    assert '20' == str(excinfo.value)
    assert 'Command "NotExist" not found.' in output(capsys)

    app.add_command(CommandContact)
    cmd = app.get_command('')
    assert cmd._name == 'help'

    app.add_command(CommandContact)
    cmd = app.get_command('--option')
    assert cmd._name == 'help'


def test_command_run(capsys):
    app = application.App()
    app.add_command(CommandContact)
    app.fire('contact NAME')
    assert 'Contact NAME ' in output(capsys)

    argv = sys.argv
    try:
        sys.argv = ['/tmp', 'contact', '-v', 'TEST_NAME']
        app = application.App()
        app.add_command(CommandContact)
        app.fire()
        printed = output(capsys)
        assert 'Contact TEST_NAME with 18 years old' in printed
        assert '"verbose" = True' in printed
    finally:
        sys.argv = argv


def test_command_argument(capsys):
    app = application.App()
    app.add_command(CommandContact)

    with pytest.raises(SystemExit) as excinfo:
        app.fire('contact')
    assert '40' == str(excinfo.value)
    assert 'The argument "name" is required' in output(capsys)

    cmd = app.get_command('contact')
    with pytest.raises(command.FieldException) as ex:
        cmd.launch('contact')
    assert 'The argument "name" is required' == str(ex.value)
    with pytest.raises(command.FieldException) as ex:
        cmd.launch('contact --option')
    assert 'The argument "name" is required' == str(ex.value)
    with pytest.raises(command.FieldException) as ex:
        cmd.launch('contact --option=option_value')
    assert 'The argument "name" is required' == str(ex.value)

    app.fire('contact NAME')
    assert 'Contact NAME' in output(capsys)
    app.fire('contact NAME --option')
    assert 'Contact NAME' in output(capsys)
    app.fire('contact NAME --option=option_value')
    assert 'Contact NAME' in output(capsys)
    app.fire('contact NAME')
    assert 'Contact NAME' in output(capsys)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME')
    assert cmd.name == 'NAME'
    assert cmd.age == 18

    cmd.parse('contact NAME 20')
    assert cmd.name == 'NAME'
    assert isinstance(cmd.age, int)
    assert cmd.age == 20

    cmd.parse('contact NAME --option=option 20')
    assert cmd.name == 'NAME'
    assert isinstance(cmd.age, int)
    assert cmd.age == 20

    cmd.parse('contact NAME --option 20')
    assert cmd.name == 'NAME'
    assert isinstance(cmd.age, int)
    assert cmd.age == 20

    cmd.parse('contact NAME --option 20 10')
    assert cmd.name == 'NAME'
    assert isinstance(cmd.age, int)
    assert cmd.age == 20


def test_command_options():
    app = application.App()
    app.add_command(CommandContact)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --option')
    assert hasattr(cmd, 'option')
    assert cmd.option is True
    assert hasattr(cmd, 'option_init')

    cmd = app.get_command('contact')
    cmd.parse('contact --str-option OPT NAME')
    assert cmd.str_option == 'OPT'

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --str OPT')
    assert cmd.str_option == 'OPT'

    cmd = app.get_command('contact')
    cmd.parse('contact --option NAME --str=OPT --int-option 99 10')
    assert cmd.option is True
    assert cmd.str_option == 'OPT'
    assert cmd.int_option == 99
    assert cmd.age == 10

    cmd = app.get_command('contact')
    cmd.parse('contact --str=OPT NAME --int 99 10')
    assert cmd.str_option == 'OPT'
    assert cmd.int_option == 99
    assert cmd.age == 10

    cmd = app.get_command('contact')
    cmd.parse('contact --option=OPT --int-option 99 NAME')
    assert cmd.option is True
    assert cmd.str_option == 'default'
    assert cmd.int_option == 99
    assert cmd.age == 18

    cmd = app.get_command('contact')
    cmd.parse('contact --str=OPT -o 99 NAME')
    assert cmd.str_option == 'OPT'
    assert cmd.int_option == 99

    cmd = app.get_command('contact')
    cmd.parse('contact --str=OPT --o 99 NAME')
    assert cmd.str_option == 'OPT'
    assert cmd.int_option == 99

    cmd = app.get_command('contact')
    cmd.parse('contact --str=OPT --o 99 NAME --not-exist-option')
    assert cmd.str_option == 'OPT'
    assert cmd.int_option == 99


def test_command_option_list():
    class CommandTest(CommandContact):
        list_option = command.Field(
            help='Sample list option',
            default=[],
            alias='list',
        )

    app = application.App()
    app.add_command(CommandTest)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --list=one,two,three')
    assert cmd.name == 'NAME'
    assert isinstance(cmd.list_option, list)
    assert cmd.list_option == ['one', 'two', 'three']


def test_command_option_list_force_type():
    class CommandTest(CommandContact):
        list_option = command.Field(
            help='Sample list option',
            default='a list',
            alias='list',
            force_type=list,
        )

    app = application.App()
    app.add_command(CommandTest)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --list=one,two,three')
    assert cmd._fields['list_option'].type == list
    assert isinstance(cmd.list_option, list)
    assert cmd.list_option == ['one', 'two', 'three']


def test_command_option_default_type():
    class CommandTest(CommandContact):
        bool_option = command.Field(
            help='Sample bool option',
        )

    app = application.App()
    app.add_command(CommandTest)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --bool-option')
    assert cmd._fields['bool_option'].type == bool
    assert isinstance(cmd.bool_option, bool)
    assert cmd.bool_option is True


def test_command_option_int_convert_error(capsys):
    app = application.App()
    app.add_command(CommandContact)

    with pytest.raises(command.FieldException):
        cmd = app.get_command('contact')
        cmd.parse('contact NAME --int ONE')

    with pytest.raises(SystemExit) as excinfo:
        app.fire('contact NAME --int ONE')
    assert '40' == str(excinfo.value)
    msg = 'The option "int_option" with he value "ONE" must be int'
    assert msg in output(capsys)


def test_command_option_alias():
    class CommandTest(CommandContact):
        bool_option = command.Field(help='Sample bool option', alias='--bool')

    app = application.App()
    app.add_command(CommandTest)

    cmd = app.get_command('contact')
    cmd.parse('contact NAME --bool')
    assert cmd.bool_option is True


def test_command_option_alias_duplicate(capsys):
    class CommandTest(CommandContact):
        bool_option = command.Field(help='Sample bool option', alias='-o')

    app = application.App()
    app.add_command(CommandTest)
    with pytest.raises(SystemExit) as excinfo:
        app.fire('contact NAME -o')
    assert '30' == str(excinfo.value)
    assert 'Duplicate option alias "o"' in output(capsys)


def test_command_option_global(capsys):
    class CommandTest(command.Command):
        _name = 'test'

        bool_option = command.Field(
            help='Sample bool option',
            alias='-o',
            default=False,
        )

        def fire(self):
            pass

    def get_test_app():
        app = application.App()
        app.add_command(CommandTest)
        app.add_option(
            'user',
            command.Field(
                help='Username for tests',
                default='ubuntu',
                alias=['-u', '--username'],
            ),
        )
        return app

    app = get_test_app()
    with pytest.raises(command.CommandException) as excinfo:
        app.add_option(
            'verbose',
            command.Field(
                help='Verbose mode',
                default=False,
                alias=['v', 'vv', 'vvv'],
            ),
        )
    assert 'Duplicate global option alias "v"' in str(excinfo.value)
    with pytest.raises(command.CommandException):
        app.add_option(
            'version',
            command.Field(
                help='Show version',
                default='1.0',
                alias=['-v'],
            ),
        )
    assert app.get_option('verbose') is False
    assert app.get_option('user') == 'ubuntu'
    assert app.get_option('not-exists', 99) == 99
    assert app.set_option('not-exists', 99) is False

    app = get_test_app()
    cmd = app.get_command('test')
    cmd.parse('test -v -o')
    assert cmd.bool_option is True
    assert app.get_option('verbose') is True
    assert app.get_option('user') == 'ubuntu'

    app = get_test_app()
    cmd = app.get_command('test')
    cmd.parse('test -v -o --user=root')
    assert cmd.bool_option is True
    assert app.get_option('verbose') is True
    assert app.get_option('user') == 'root'

    app = get_test_app()
    app.add_command(CommandTest)
    cmd = app.get_command('test')
    cmd.parse('test -v -o --user root')
    assert cmd.bool_option is True
    assert app.get_option('verbose') is True
    assert app.get_option('user') == 'root'

    output(capsys)
    app = get_test_app()
    app.add_command(CommandTest)
    assert app.get_option('verbose') is False
    app.fire('test -o')
    assert app.get_option('verbose') is False
    assert 'Global option "verbose" = True' not in output(capsys)
    app.fire('test -v -o')
    assert app.get_option('verbose') is True
    assert 'Global option "verbose" = True' in output(capsys)


def test_command_extra_args(capsys):
    app = application.App()
    app.add_command(CommandContact)

    cmd = app.get_command('contact')
    assert cmd.extra_args == []
    cmd.parse('contact Rob 21')
    assert cmd.extra_args == []

    cmd = app.get_command('contact')
    cmd.parse('contact Rob 21 extra args')
    assert cmd.extra_args == ['extra', 'args']

    cmd = app.get_command('contact')
    cmd.parse('contact Rob 21 --int 1 --extra args')
    assert cmd.extra_args == ['--extra', 'args']

    cmd = app.get_command('contact')
    cmd.parse('contact Rob 21 --extra args --int 1')
    assert cmd.extra_args == ['--extra', 'args']
    assert cmd.int_option == 1
    cmd.parse('contact Rob 21')
    assert cmd.int_option == 1

    cmd = app.get_command('contact')
    cmd.parse('contact Rob 21 --extra e1 --int 1 e2 e3 --extra-option o1 o2')
    assert cmd.extra_args == [
        '--extra',
        'e1',
        'e2',
        'e3',
        '--extra-option',
        'o1',
        'o2',
    ]


def test_command_return_error_code(capsys):
    class CommandTest(command.Command):
        _name = 'test'

        def fire(self):
            return 99

    app = application.App()
    app.add_command(CommandTest)
    with pytest.raises(SystemExit) as excinfo:
        app.fire('test')
    assert excinfo.value.code == 99



import os
import re
from typing import List

import jinja2
from clifire import application


class Template:
    def __init__(self, template_folder: str):
        self.template_folder = template_folder
        self.jinja2 = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_folder)
        )

    def path(self, *args: List[str]) -> str:
        return application.App.current_app.path(self.template_folder, *args)

    def render(self, template, **args):
        args['os'] = os
        return self.jinja2.get_template(template).render(**args)

    def write(self, template, filename, mark=None, **args):
        template_content = self.render(template, **args)

        if not mark:
            with open(filename, 'w') as file:
                file.write(template_content)
            return template_content

        content = ''
        if os.path.exists(filename):
            with open(filename) as file:
                content = file.read()
        mark_escaped = re.escape(mark)
        pattern = f'({mark_escaped})[\\s\\S]*?\\1'
        if re.search(pattern, content):
            content = re.sub(
                pattern, f'{mark}\n{template_content}\n{mark}', content
            )
        else:
            content += f'\n{mark}\n{template_content}\n{mark}\n'
        with open(filename, 'w') as file:
            file.write(content)
        return content



from clifire import result


def test_result():
    res = result.Result(0, 'OK', 'KO')
    assert bool(res) is True
    assert res.code == 0
    assert str(res)
    assert str(res) == res.__repr__()

    res = result.ResultOk('OK')
    assert bool(res) is True
    assert res.code == 0
    assert res.stdout == 'OK'
    assert res.stderr == ''

    res = result.ResultError('ERROR', 99)
    assert bool(res) is False
    assert res.code == 99
    assert res.stdout == ''
    assert res.stderr == 'ERROR'



import os
import subprocess
import sys
import time
from unittest.mock import patch

import pytest
from clifire import application, command, out


def output(capsys):
    captured = capsys.readouterr()
    return out.ansi_clean(captured.out)


def test_app_keyboard_interrupt(capsys):
    class CommandWait(command.Command):
        _name = 'wait'

        def fire(self):
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                out.info('KeyboardInterrupt received')
                raise

    app = application.App()
    app.add_command(CommandWait)

    with patch('time.sleep', side_effect=KeyboardInterrupt):
        with pytest.raises(KeyboardInterrupt):
            app.fire('wait')

    printed = output(capsys)
    assert 'KeyboardInterrupt received' in printed
    assert 'End command' not in printed
    assert 'Keyboard interrupt!' in printed


def test_app_global_option_no_ansi(capsys):
    app = application.App(name='sample', version='1.0.99')
    assert app.get_option('no_ansi') is False
    app.fire('help -v')
    assert 'Console output setup with ANSI' in output(capsys)

    app.fire('help --no-ansi -v')
    assert 'Console output setup with no ANSI' in output(capsys)
    assert app.get_option('no_ansi') is True

    argv = sys.argv
    try:
        sys.argv = ['/tmp', 'help', '--no-ansi']
        app = application.App()
        assert app.get_option('no_ansi') is True
    finally:
        sys.argv = argv


def test_app_no_command_provided(capsys):
    app = application.App(command_help=None)
    with pytest.raises(SystemExit) as excinfo:
        app.fire('')
    assert '10' == str(excinfo.value)
    assert 'No command provided.' in output(capsys)


def test_shell():
    app = application.App()
    res = app.shell('pwd')
    assert res.stdout in __file__

    test_path = os.path.dirname(__file__)
    assert os.getcwd() != test_path
    res = app.shell('pwd', path=test_path)
    assert res.stdout == test_path

    res = app.shell('echo $MY_ENV_VALUE', env={'MY_ENV_VALUE': 'is_ok!'})
    assert 'is_ok!' in res.stdout

    res = app.shell('this command not exist!')
    assert not res.stdout
    assert res.stderr

    pwd = os.getcwd()
    res = app.shell('pwd')
    assert res.stdout == pwd
    res = app.shell('pwd', path=app.path(test_path, 'sample'))
    assert pwd == os.getcwd()
    assert res.stdout != pwd
    assert 'tests/sample' in res.stdout
    res = app.shell('pwd')
    assert res.stdout == pwd
    assert 'tests/sample' not in res.stdout


def test_shell_called_process_error():
    app = application.App()
    with patch('subprocess.run') as mocked_run:
        error = subprocess.CalledProcessError(
            returncode=1, cmd='dummy', stderr=b'error occurred'
        )
        mocked_run.side_effect = error

        res = app.shell('dummy')
        assert isinstance(res, application.result.ResultError)
        assert res.code == 1
        assert 'error occurred' in res.stderr


def test_path():
    assert application.App.path('/tmp') == '/tmp'

    app = application.App()
    assert app.path('/tmp') == '/tmp'
    assert app.path('/tmp', 'test', 'file.txt') == '/tmp/test/file.txt'

    home = os.path.expanduser('~')
    assert app.path('~/test.txt') == os.path.join(home, 'test.txt')
    assert app.path('./test') == os.path.abspath('./test')
    assert app.path('/tmp/my folder/file.txt') == '/tmp/my folder/file.txt'
    assert app.path('~/folder1', 'folder2', 'file.txt') == os.path.join(
        home, 'folder1', 'folder2', 'file.txt'
    )

    assert app.path() == os.getcwd()



import importlib.util
import os

from clifire import application, out


def load(path):
    if not os.path.exists(path):
        out.debug2('Not exist, skipping')
        return False
    if os.path.isdir(path):
        load_folder(path)
    else:
        load_file(path)
    return True


def load_folder(path):
    for file in os.listdir(path):
        if file.endswith('.py'):
            load_file(os.path.join(path, file))


def load_file(filename):
    out.debug2(f'Loading {os.path.relpath(filename)}')
    module_name = os.path.basename(filename)[:-3]
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def main(command_line: str = None):
    app = application.App(name='CliFire', version='0.1.11')
    current_dir = os.getcwd()
    out.debug(f'Search commands in {current_dir} folder and parents')
    loaded = False
    while True:
        fire_folder = os.path.join(current_dir, 'fire')
        if load(fire_folder):
            loaded = True
            break
        fire_file = os.path.join(current_dir, 'fire.py')
        if load(fire_file):
            loaded = True
            break
        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:
            break
        current_dir = parent_dir
    if not loaded:
        out.warn(
            'The file fire.py or folder fire is not in this directory or its '
            'parents'
        )
        out.critical('Fire not found!')
    app.fire(command_line)


if __name__ == '__main__':
    main()



import os
import stat

from clifire import command, out


@command.fire
def install(cmd, _global: False):
    '''
    Install global `fire` command

    Args:
    global: Install the command globally
    '''
    install_path = cmd.app.path('~/.local/bin')
    if not os.path.exists(install_path):
        out.debug2(f'Install path {install_path} not exist')
        install_path = cmd.app.path('/usr/local/bin')
    if not os.path.exists(install_path):
        out.debug2(f'Install path {install_path} not exist')
        out.critical('No path for install')
    fire_path = f'{install_path}/fire'
    out.info(f'Command available at {fire_path}')
    pyproject_path = cmd.app.path('pyproject.toml')
    with open(fire_path, 'w') as fp:
        fp.write(
            '\n'.join(
                [
                    '#!/bin/bash',
                    f'rye run --pyproject {pyproject_path} fire "$@"',
                ]
            )
        )
    os.chmod(
        fire_path,
        stat.S_IRWXU
        | stat.S_IRGRP
        | stat.S_IXGRP
        | stat.S_IROTH
        | stat.S_IXOTH,
    )
    out.success('Installation completed!')


@command.fire
def build(cmd):
    '''
    Build the package and documentation
    '''
    live = out.LiveText('Building ...')
    res = cmd.app.shell('rye build')
    if res:
        live.success('Build success')
    else:
        live.error('Error on building packages.')
        out.error(res.stderr)
        return 1

    live.info('Building the docs ..')
    cmd.app.fire('doc build')


@command.fire
def publish(cmd):
    '''
    Publish then package
    '''
    res = cmd.app.shell('rye publish -y', capture_output=False)
    if res:
        out.success('Published')
    else:
        out.critical('Error on publish package')


@command.fire
def coverage(cmd, file: str = ''):
    '''
    Launch tests with coverage

    Args:
        file: Specific test file to run (e.g., "test_command.py")
    '''
    if file:
        test_path = f'tests/{file}' if not file.startswith('tests/') else file
        if not os.path.exists(test_path):
            out.critical(f'File {test_path} not exists')
        bash = (
            f'rye run coverage run -m pytest {test_path} '
            '&& rye run coverage html'
        )
    else:
        bash = 'rye run coverage run -m pytest && rye run coverage html'

    cmd.app.shell(bash, capture_output=False)


@command.fire
def precommit(cmd):
    '''
    Launch pre-commit
    '''
    cmd.app.shell('rye run pre-commit run --all-files', capture_output=False)


@command.fire
def tests(cmd):
    '''
    Launch tests
    '''
    path = cmd.app.path(os.path.dirname(__file__), '..')
    cmd.app.shell('rye run python -m pytest', capture_output=False, path=path)



from clifire import application, command, out


class HelloCommand(command.Command):
    _name = 'hello'
    _help = 'Create template'

    title = command.Field(
        pos=1,
        force_type=str,
    )

    def fire(self):
        print(self.title)
        content = self.app.template.render(
            'hello.jinja2', title=self.title, numbers=[1, 2, 3]
        )
        out.success(content)


def main():
    app = application.App(template_folder='./sample_template')
    app.add_command(HelloCommand)
    app.fire()


if __name__ == '__main__':
    main()



from clifire import command, out


@command.fire
def hello(cmd, user: str = '', _sudo: bool = False):
    '''
    Display a greeting on the console.

    Args:
        user: Name of the user to greet. If empty, the current system user is used.
        _sudo: Run the command with sudo privileges.
    '''
    if not user:
        sudo = 'sudo' if _sudo else ''
        user = cmd.app.shell(f'{sudo} whoami').stdout
    out.info(f'Hi {user}!')
